/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "NSXToolController.h"
#include "controllers/common/family/Families.h"
#include "rapidjson/document.h"
#include "rapidjson/rapidjson.h"
#include "rapidjson/error/error.h"
#include "rapidjson/error/en.h"
#include <boost/lexical_cast.hpp>
#include <common/base/Date.h>
#include <fstream>
#include <iostream>
#include <nexus/napi.h>

namespace nsxtool {

using namespace std;
using namespace common;
using namespace boost;
using namespace cameo;
using namespace rapidjson;

const std::string NSXToolController::TYPE = "nsxtool_controller";
const std::string NSXToolController::REMOTE_APPLICATION = "nsx-plotter-server";
const std::string NSXToolController::RESPONDER = "responder";
const std::string NSXToolController::NSX_PUBLISHER = "nsx_plotter_pub";

NSXToolController::NSXToolController(const std::string& name) :
		ExperimentController(name), controller::Stoppable(this) {

	setFamily(family::ACQUISITION, family::COUNT);

	serverEndpoint.init(this, SAVE, "cameo_server");
	initialized.init(this, NOSAVE, "initialized");
	fileDirectory.init(this, SAVE, "file_directory");
	filename.init(this, SAVE, "filename");

	x.init(this, NOSAVE, "x");
	y.init(this, NOSAVE, "y");
	data.init(this, NOSAVE, "data");
}

NSXToolController::NSXToolController(const NSXToolController& controller) :
		ExperimentController(controller), controller::Stoppable(this) {
}

NSXToolController::~NSXToolController() {

	// Test the subscriber
	if (m_subscriber.get() != 0) {

		// Stop the subscriber
		m_subscriber->cancel();

		// The subscriber thread terminates
		m_subscriberThread->join();
	}

	if (m_remoteApplication.get() != 0) {
		// Stop the remote application.
		m_remoteApplication->stop();

		// Wait for the termination
		application::State state = m_remoteApplication->waitFor();

		cout << "Remote application " << m_remoteApplication->getName() << " terminated with state " << application::toString(state) << endl;
	}
}

void NSXToolController::postConfiguration() {

	if (serverEndpoint() == "") {
		m_server.reset(new Server(application::This::getServer().getEndpoint()));

	} else {
		m_server.reset(new Server(serverEndpoint()));
	}

	initialized = initApplication();
}

bool NSXToolController::initApplication() {

	// Do not initialize if it is already done
	if (initialized()) {
		return true;
	}

	// Start the remote server
	if (!m_server->isAvailable(1000)) {
		cout << "Remote server is not available" << endl;
		return false;
	}

	cout << "Remote server is connected" << endl;

	m_remoteApplication = m_server->connect(REMOTE_APPLICATION);
	if (m_remoteApplication->exists()) {
		// The application exists from a previous server session
		m_remoteApplication->kill();
		application::State state = m_remoteApplication->waitFor();
		cout << "Terminated remote application with state " << application::toString(state) << endl;
	}

	m_remoteApplication = m_server->start(REMOTE_APPLICATION);

	if (!m_remoteApplication->exists()) {
		cout << "No remote application" << endl;
		return false;
	}

	// Create the subscriber after the publisher
	m_subscriber = application::Subscriber::create(*m_remoteApplication, NSX_PUBLISHER);
	cout << "subscriber " << *m_subscriber << endl;

	// Start the subscriber loop
	m_subscriberThread.reset(new boost::thread(boost::bind(&NSXToolController::subscriberLoop, this)));

	// Application initialized
	initialized = true;

	return true;
}

void NSXToolController::readFile(const std::string& nexusFileName, std::string& fileContent) {

	streampos size;

	ifstream file(nexusFileName.c_str(), ios::in | ios::binary | ios::ate);

	if (file.is_open()) {

		size = file.tellg();

		// Reserve the string.
		fileContent.resize(size);

		// Get the allocated array.
		char * array = (char *) fileContent.data();

		file.seekg(0, ios::beg);
		file.read(array, size);
		file.close();

	} else {
		cerr << "File " << nexusFileName << " cannot be read." << endl;
	}
}

void NSXToolController::subscriberLoop() {

	// Loop on events
	string data;
	while (m_subscriber->receive(data)) {
		readProcessedData(data);
	}
}

void NSXToolController::readProcessedData(const std::string& response) {

	//string test = "[[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0], [9.0, 10.0, 11.0, 12.0]]";
	//cout << "response " << endl << response << endl;

	Document document;

	ParseResult res = document.Parse(response.c_str());
	if (!res) {
		cerr << "Error while parsing NSXTool response : " <<  GetParseError_En(res.Code()) << "(" << res.Offset() << ")" << endl;
		return;
	}

	// Check array.
	if (!document.IsArray()) {
		cerr << "Error while parsing NSXTool response : bad format" << endl;
		return;
	}

	// Get the array.
	Document::Array array = document.GetArray();

	// Check y size.
	int ySize = array.Size();
	if (ySize == 0) {
		return;
	}

	// We resize the y data.
	resizeData(y, ySize);

	// Check array.
	if (!array[0].IsArray()) {
		cerr << "Error while parsing NSXTool response : bad format" << endl;
		return;
	}

	// Check that all the lines have the same number of elements.
	int xSize = array[0].GetArray().Size();

	// Iterate the lines.
	for (int i = 1; i < ySize; ++i) {

		// Check array.
		if (!array[i].IsArray()) {
			cerr << "Error while parsing NSXTool response : bad format" << endl;
			return;
		}

		Document::Array lineArray = array[i].GetArray();

		if (lineArray.Size() != xSize) {
			cerr << "Bad number of elements when reading response from NSXTool" << endl;
			return;
		}
	}

	// We can resize the x data and the z data.
	resizeData(x, xSize);
	resizeData(data, xSize * ySize);

	// Reading the data by iterating the lines.
	for (int i = 0; i < ySize; ++i) {

		Document::Array lineArray = array[i].GetArray();

		// Copy the elements.
		for (int j = 0; j < xSize; ++j) {
			data()[i * xSize + j] = lineArray[j].GetDouble();
		}
	}

	x.sendEvent();
	y.sendEvent();
	data.sendEvent();

	cout << "Processing finished." << endl;
}

void NSXToolController::start() {

	if (!initApplication()) {
		return;
	}

	// Send a simple message with the content of the Nexus data file.
	string fileContent;

	// Read the file.
	readFile(fileDirectory() + "/" + filename(), fileContent);

	// Create the requester
	unique_ptr<application::Requester> requester = application::Requester::create(*m_remoteApplication, RESPONDER);

	if (requester.get() == 0) {
		cout << "requester error" << endl;
		return;
	}

	// Start date.
	Date begin;

	// Send the file content to the server.
	requester->sendBinary(fileContent);

	// Wait for the response from the server.
	string response;
	requester->receive(response);

	// Print the duration.
	Date end;
	double ms = (end - begin).getMilliseconds();
	cout << "File sent and received in " << ms << " ms" << endl;
}

void NSXToolController::stop() {

}

void NSXToolController::resizeData(ArrayProperty<float64> & property, int size) {

	if (size != property.getSize()) {

		delete [] property.get();
		double * values = new double[size];

		for (int i = 0; i < size; ++i) {
			values[i] = static_cast<float64>(i);
		}

		property.set(values);
		property.setSize(size);
	}
}

}
