/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "RemoteDPPCoincidence.h"
#include <common/base/ServerProperties.h>
#include <boost/thread.hpp>
#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/property_tree/json_parser.hpp>
#include <cstdlib>

namespace npp {

using namespace std;
using namespace common;
using namespace boost;
using namespace cameo;
using namespace lstdpp128;

using boost::property_tree::ptree;

const std::string RemoteDPPCoincidence::TYPE = "remote_dpp_coincidence";

RemoteDPPCoincidence::RemoteDPPCoincidence(const std::string& name) :
	RemoteDPPLiveProcess(name) {

	dT.init(this, SAVE, "dT");
	cleanDT.init(this, SAVE, "clean_dT");
	numberOfDetectorChannels.init(this, SAVE, "number_of_detector_channels");
	detectorChannel.init(this, SAVE, "detector_channel");
	coincidenceResolution.init(this, SAVE, "coincidence_resolution");
	maxBlock.init(this, SAVE, "max_block");
	totalNumberOfChannels.init(this, NOSAVE, "total_number_of_channels");

	nOrder.init(this, NOSAVE, "n_order");
	order.init(this, NOSAVE, "order");
	globalRate.init(this, NOSAVE, "global_rate");
	globalCleanRate.init(this, NOSAVE, "global_clean_rate");
	detectorRate.init(this, NOSAVE, "detector_rate");
	detectorCleanRate.init(this, NOSAVE, "detector_clean_rate");

	crateId.init(this, NOSAVE, "crate_id");
	boardId.init(this, NOSAVE, "board_id");
	channelId.init(this, NOSAVE, "channel_id");
	channelRate.init(this, NOSAVE, "channel_rate");

	registerRefresher(numberOfDetectorChannels, &RemoteDPPCoincidence::refreshNumberOfDetectorChannels, this);
}

RemoteDPPCoincidence::RemoteDPPCoincidence(const RemoteDPPCoincidence& controller) :
	RemoteDPPLiveProcess(controller) {
}

RemoteDPPCoincidence::~RemoteDPPCoincidence() {
}

void RemoteDPPCoincidence::postConfiguration() {

	RemoteDPPLiveProcess::postConfiguration();

	// order
	nOrder.set(7);

	order.resize(7);
	globalRate.resize(7);
	globalCleanRate.resize(7);
	detectorRate.resize(7);
	detectorCleanRate.resize(7);

	// properties
	for (int32 i = 0; i < 7; i++) {
		order.set(i, i + 2);
	}

	resetData();
}

void RemoteDPPCoincidence::setDetectorControllers(const std::vector<ControllerPtr<acquisition::DetectorController> > & detectorControllers) {

	DPPLiveProcess::setDetectorControllers(detectorControllers);

	/* Do not do it
	// register the list mode data
	int32 index = 0;
	for (vector<ControllerPtr<acquisition::DetectorController> >::const_iterator c = detectorControllers.begin(); c != detectorControllers.end(); ++c) {
		registerUpdater((*c)->listModeData, &RemoteDPPCoincidence::updateListModeData, this, index);
		index++;
	}
	*/
}

void RemoteDPPCoincidence::start() {

	// Need a mutex to avoid having two start at the same time.
	{
		boost::mutex::scoped_lock lock(m_mutex);

		initFileReading();
		resetData();

		// Numor
		string numor;
		size_t size = fileName().size();

		if (size < 6) {
			string prefix(6 - size, '0');
			numor = prefix + fileName();
		} else {
			numor = fileName();
		}

		// Setting the parameters
		ptree parameters;

		parameters.put("numor", numor);
		parameters.put("dT", dT() / 10);
		parameters.put("cleanDT", cleanDT() / 10);

		// detector channels
		ptree detectorChannelArray;
		ptree channel;
		int32 detectorChannelSize = detectorChannel.getSize();
		for (int32 i = 0; i < detectorChannelSize; ++i) {
			channel.put("", detectorChannel(i));
			detectorChannelArray.push_back(std::make_pair("", channel));
		}
		parameters.put_child("detectorChannels", detectorChannelArray);

		parameters.put("matrixResolution", coincidenceResolution());
		parameters.put("maxBlocks", maxBlock());
		parameters.put("blockSize", blockSize());
		parameters.put("refreshTime", refreshTimeS());

		ptree context = writeListModeContextToPropertyTree(listModeContext);

		parameters.put_child("context", context);

		// Serialize the parameters
		stringstream os;
		write_json(os, parameters, false);


		// Kill the application if one remains
		m_server->killAllAndWaitFor("ncoincidence");

		vector<string> args;
		args.push_back(os.str());

		// Start the application
		m_coincidenceApplication = m_server->start("ncoincidence", args);

		// Test if application started
		if (!m_coincidenceApplication->exists()) {
			log(Level::s_Error) << "cannot start coincidence application" << endlog;
			return;
		}

		// Create the publisher after the start of the application, because it is a blocking call
		m_publisher.reset();
		m_publisher = application::Publisher::create("dpp_publisher", 1);

		cout << "publisher " << *m_publisher << endl;

		// Wait for the subscriber
		bool sync = m_publisher->waitForSubscribers();

		// Create the subscriber after the publisher
		unique_ptr<cameo::application::Subscriber> subscriber = application::Subscriber::create(*m_coincidenceApplication, "dpp_coincidence_results");
		cout << "subscriber " << *subscriber << endl;

		// Start the test thread if necessary
		unique_ptr<boost::thread> thread = startFileReading();

		// Listen to the result stream
		vector<double> channelRates;
		while (subscriber->receive(channelRates)) {
			// copying the rates
			for (int32 i = 0; i < totalNumberOfChannels(); i++) {
				channelRate.set(i, channelRates[i]);
			}

			vector<double> coincidenceRates;
			if (!subscriber->receive(coincidenceRates)) {
				break;
			}

			if (coincidenceRates.size() <= 4 * 7) {
				// Copying values
				for (int32 i = 0; i < 7; ++i) {
					globalRate.set(i, coincidenceRates[i]);
					globalCleanRate.set(i, coincidenceRates[7 + i]);
					detectorRate.set(i, coincidenceRates[7 * 2 + i]);
					detectorCleanRate.set(i, coincidenceRates[7 * 3 + i]);
				}

			} else {
				cerr << "problem with coincidence rates size" << endl;
			}
		}

		// Wait for the termination of the application
		application::State state = m_coincidenceApplication->waitFor();

		// Join the test thread
		finishFileReading(thread.get());

		cout << "coincidence application terminated with state " << application::toString(state) << endl;
	}
}

void RemoteDPPCoincidence::raz() {
	resetData();
}

void RemoteDPPCoincidence::refreshNumberOfDetectorChannels(int32 value) {
	detectorChannel.resize(value);
}

void RemoteDPPCoincidence::updateListModeData(int index) {

	if (m_publisher.get() != 0 && !m_publisher->hasEnded()) {
		int32 size = detectorControllers[index]->listModeData.getSize();
		int32 * memoryBlock = detectorControllers[index]->listModeData();

		// put the array into a string
		string data(reinterpret_cast<char *>(memoryBlock), sizeof(int32) * size);

		// send the data
		m_publisher->send(data);

		cout << "updateListModeData " << index << " size " << size << endl;
	}
}

void RemoteDPPCoincidence::resetData() {

	// Calculate the total number of channels
	BoardIterator b;

	if (b.end()) {
		totalNumberOfChannels = 0;

	} else {
		int32_t currentNbChannels = (*b).nbChannels;

		++b;
		while (!b.end()) {
			currentNbChannels += (*b).nbChannels;
			++b;
		}

		totalNumberOfChannels = currentNbChannels;
	}

	cout << "totalNumberOfChannels " << totalNumberOfChannels() << endl;

	// resetting values
	for (int32 i = 0; i < 7; ++i) {
		globalRate.set(i, 0.0);
		globalCleanRate.set(i, 0.0);
		detectorRate.set(i, 0.0);
		detectorCleanRate.set(i, 0.0);
	}

	crateId.resize(totalNumberOfChannels());
	boardId.resize(totalNumberOfChannels());
	channelId.resize(totalNumberOfChannels());
	channelRate.resize(totalNumberOfChannels());

	int32 i = 0;

	// Set the ids
	BoardIterator bc;
	while (!bc.end()) {

		if (boardHasEnergy((*bc).boardType)) {

			int32 nbChannels = (*bc).nbChannels;
			int32 histogramSize = (1 << boardEnergyResolution((*bc).boardType));

			for (int32 j = 0; j < nbChannels; ++j) {
				crateId.set(i, bc.crate());
				boardId.set(i, bc.board());
				channelId.set(i, j);
				channelRate.set(i, 0.0);

				++i;
			}
		}

		++bc;
	}
}

}
