/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "CAENSy2527Driver.h"

#include "drivers/global/DriversCommands.h"
#include "CAENSy2527Def.h"
#include "CAENSy2527State.h"
#include "PerfectCAENSy2527Driver.h"
#include "RealCAENSy2527Driver.h"

using namespace std;

namespace caen_sy2527 {

const std::string CAENSy2527Driver::TYPE = "caen_sy2527";

/*
 * Constructor
 */
CAENSy2527Driver::CAENSy2527Driver(const string& name) {

	driver::SocketTcp::init(name);

	registerStates(new RealCAENSy2527Driver(this), new PerfectCAENSy2527Driver(this), new PerfectCAENSy2527Driver(this));

	// Crate properties
	crateModel.init(this, NOSAVE, "crate_model");
	linkType.init(this, NOSAVE, "link_type");
	nbSlots.init(this, NOSAVE, "nb_slots");
	channelsPerBoard.init(this, NOSAVE, "channels_per_board");
	boardModel.init(this, NOSAVE, "board_model");
	boardDescription.init(this, NOSAVE, "board_description");
	boardSerial.init(this, NOSAVE, "board_serial");
	boardFirmware.init(this, NOSAVE, "board_firmware");

	// Init the device command list
	initCommand(driver::INIT_COMMAND);
	initCommand(driver::READ_INFOS_COMMAND);

	// Init functions
	registerFunction(CAENSY2527_FUNCTION);

	deviceType = BRANCH_DEVICE_TYPE_DEVICE_CONTAINER;
	numberLeaf = CAENSY2527_MAX_BOARDS;
//	deviceType = LEAF_DEVICE_TYPE_DEVICE_CONTAINER;

	m_crateHandle = -1;
}

/*
 * Destructor
 */
CAENSy2527Driver::~CAENSy2527Driver() {

}

/*
 * execute
 */
void CAENSy2527Driver::execute(const std::string& aCommand) {

	if ((aCommand != driver::STOP_COMMAND) &&
			(aCommand != driver::STATUS_COMMAND) &&
			(aCommand != driver::READ_COMMAND)) {
		sendProgressEvent(PROGRESSION_UNKNOWNSTATE_DEVICE_CONTAINER);
	}

	CAENSy2527State* currentState = dynamic_cast<CAENSy2527State*>(getCurrentState());
	// Check command
	if (aCommand == driver::INIT_COMMAND) {
		// Init command
		currentState->init();
	} else if (aCommand == driver::READ_INFOS_COMMAND) {
		// Info command
		currentState->readInfos();
		commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
	} else {
		//Error bad command
		//Todo , normaly it will be detect it device action driver
	}
}

/*
 * read
 */
void CAENSy2527Driver::read(CAENSy2527BoardDriver* aBoard) {

	CAENSy2527State* currentState = static_cast<CAENSy2527State*> (getCurrentState());
	currentState->read(aBoard);
}

/*
 * readParam
 */
void CAENSy2527Driver::readParam(CAENSy2527BoardDriver *aBoard) {

	CAENSy2527State* currentState = static_cast<CAENSy2527State*> (getCurrentState());
	currentState->readParam(aBoard);
}

/*
 * writeParam
 */
void CAENSy2527Driver::writeParam(CAENSy2527BoardDriver *aBoard) {

	CAENSy2527State* currentState = static_cast<CAENSy2527State*> (getCurrentState());
	currentState->writeParam(aBoard);
}

/*
 * start
 */
void CAENSy2527Driver::start(CAENSy2527BoardDriver *aBoard) {

	CAENSy2527State* currentState = static_cast<CAENSy2527State*> (getCurrentState());
	currentState->start(aBoard);
}

/*
 * stop
 */
void CAENSy2527Driver::stop(CAENSy2527BoardDriver *aBoard) {

	CAENSy2527State* currentState = static_cast<CAENSy2527State*> (getCurrentState());
	currentState->stop(aBoard);
}

/*
 * readStatus
 */
void CAENSy2527Driver::readStatus(CAENSy2527BoardDriver *aBoard) {

	CAENSy2527State* currentState = static_cast<CAENSy2527State*> (getCurrentState());
	currentState->readStatus(aBoard);
}

/*
 * powerOn
 */
void CAENSy2527Driver::powerOn(CAENSy2527BoardDriver *aBoard) {

	CAENSy2527State* currentState = static_cast<CAENSy2527State*> (getCurrentState());
	currentState->powerOn(aBoard);
}

/*
 * powerOff
 */
void CAENSy2527Driver::powerOff(CAENSy2527BoardDriver *aBoard) {

	CAENSy2527State* currentState = static_cast<CAENSy2527State*> (getCurrentState());
	currentState->powerOff(aBoard);
}

}
